@@ -5,6 +5,7 @@
 
 use super::{exp_acc, Digest, ElementHasher, Hasher};
 use core::convert::TryInto;
+use core::ops::Range;
 use math::{fields::f64::BaseElement, FieldElement, StarkField};
 
 mod digest;
@@ -19,10 +20,23 @@ mod tests;
 /// Sponge state is set to 12 field elements or 768 bytes; 8 elements are reserved for rate and
 /// the remaining 4 elements are reserved for capacity.
 const STATE_WIDTH: usize = 12;
-const RATE_WIDTH: usize = 8;
+
+/// The rate portion of the state is located in elements 4 through 11.
+const RATE_RANGE: Range<usize> = 4..12;
+const RATE_WIDTH: usize = RATE_RANGE.end - RATE_RANGE.start;
+
+const INPUT1_RANGE: Range<usize> = 4..8;
+const INPUT2_RANGE: Range<usize> = 8..12;
+
+/// The capacity portion of the state is located in elements 0, 1, 2, and 3.
+const CAPACITY_RANGE: Range<usize> = 0..4;
 
 /// The output of the hash function is a digest which consists of 4 field elements or 32 bytes.
-const DIGEST_SIZE: usize = 4;
+///
+/// The digest is returned from state elements 4, 5, 6, and 7 (the first four elements of the
+/// rate portion).
+const DIGEST_RANGE: Range<usize> = 4..8;
+const DIGEST_SIZE: usize = DIGEST_RANGE.end - DIGEST_RANGE.start;
 
 /// The number of rounds is set to 7 to target 128-bit security level with 40% security margin;
 /// computed using algorithm 7 from <https://eprint.iacr.org/2020/1143.pdf>
@@ -49,8 +63,13 @@ const INV_ALPHA: u64 = 10540996611094048183;
 ///   margin used in the specifications (a 50% margin rounds up to 8 rounds). The primary
 ///   motivation for this is that having the number of rounds be one less than a power of two
 ///   simplifies AIR design for computations involving the hash function.
+/// * We use the first 4 elements of the state (rather than the last 4 elements of the state) for
+///   capacity and the remaining 8 elements for rate. The output of the hash function comes from
+///   the first four elements of the rate portion of the state (elements 4, 5, 6, and 7). This
+///   effectively applies a fixed bit permutation before and after XLIX permutation. We assert
+///   without proof that this does not affect security of the construction.
 /// * When hashing a sequence of elements, we do not append Fp(1) followed by Fp(0) elements
-///   to the end of the sequence as padding. Instead, we initialize one of the capacity elements
+///   to the end of the sequence as padding. Instead, we initialize the first capacity element
 ///   to the number of elements to be hashed, and pad the sequence with Fp(0) elements only. This
 ///   ensures consistency of hash outputs between different hashing methods (see section below).
 ///   However, it also means that our instantiation of Rescue Prime cannot be used in a stream
@@ -102,11 +121,11 @@ impl Hasher for Rp64_256 {
             bytes.len() / 7 + 1
         };
 
-        // initialize state to all zeros, except for the last element of the capacity part, which
+        // initialize state to all zeros, except for the first element of the capacity part, which
         // is set to the number of elements to be hashed. this is done so that adding zero elements
         // at the end of the list always results in a different hash.
         let mut state = [BaseElement::ZERO; STATE_WIDTH];
-        state[STATE_WIDTH - 1] = BaseElement::new(num_elements as u64);
+        state[CAPACITY_RANGE.start] = BaseElement::new(num_elements as u64);
 
         // break the string into 7-byte chunks, convert each chunk into a field element, and
         // absorb the element into the rate portion of the state. we use 7-byte chunks because
@@ -130,7 +149,7 @@ impl Hasher for Rp64_256 {
             // convert the bytes into a field element and absorb it into the rate portion of the
             // state; if the rate is filled up, apply the Rescue permutation and start absorbing
             // again from zero index.
-            state[i] += BaseElement::new(u64::from_le_bytes(buf));
+            state[RATE_RANGE.start + i] += BaseElement::new(u64::from_le_bytes(buf));
             i += 1;
             if i % RATE_WIDTH == 0 {
                 Self::apply_permutation(&mut state);
@@ -147,43 +166,43 @@ impl Hasher for Rp64_256 {
         }
 
         // return the first 4 elements of the state as hash result
-        ElementDigest::new(state[..DIGEST_SIZE].try_into().unwrap())
+        ElementDigest::new(state[DIGEST_RANGE].try_into().unwrap())
     }
 
     fn merge(values: &[Self::Digest; 2]) -> Self::Digest {
         // initialize the state by copying the digest elements into the rate portion of the state
-        // (8 total elements), and set the last capacity element to 8 (the number of elements to
+        // (8 total elements), and set the first capacity element to 8 (the number of elements to
         // be hashed).
         let mut state = [BaseElement::ZERO; STATE_WIDTH];
-        state[..RATE_WIDTH].copy_from_slice(Self::Digest::digests_as_elements(values));
-        state[STATE_WIDTH - 1] = BaseElement::new(RATE_WIDTH as u64);
+        state[RATE_RANGE].copy_from_slice(Self::Digest::digests_as_elements(values));
+        state[CAPACITY_RANGE.start] = BaseElement::new(RATE_WIDTH as u64);
 
         // apply the Rescue permutation and return the first four elements of the state
         Self::apply_permutation(&mut state);
-        ElementDigest::new(state[..DIGEST_SIZE].try_into().unwrap())
+        ElementDigest::new(state[DIGEST_RANGE].try_into().unwrap())
     }
 
     fn merge_with_int(seed: Self::Digest, value: u64) -> Self::Digest {
         // initialize the state as follows:
-        // - seed is copied into the first 4 elements of the state.
-        // - if the value fits into a single field element, copy it into the fifth state element
-        //   and set the last capacity element to 5 (the number of elements to be hashed).
+        // - seed is copied into the first 4 elements of the rate portion of the state.
+        // - if the value fits into a single field element, copy it into the fifth rate element
+        //   and set the first capacity element to 5 (the number of elements to be hashed).
         // - if the value doesn't fit into a single field element, split it into two field
-        //   elements, copy them into state elements 5 and 6, and set the last capacity element
+        //   elements, copy them into rate elements 5 and 6, and set the first capacity element
         //   to 6.
         let mut state = [BaseElement::ZERO; STATE_WIDTH];
-        state[..DIGEST_SIZE].copy_from_slice(seed.as_elements());
-        state[DIGEST_SIZE] = BaseElement::new(value);
+        state[INPUT1_RANGE].copy_from_slice(seed.as_elements());
+        state[INPUT2_RANGE.start] = BaseElement::new(value);
         if value < BaseElement::MODULUS {
-            state[STATE_WIDTH - 1] = BaseElement::new(DIGEST_SIZE as u64 + 1);
+            state[CAPACITY_RANGE.start] = BaseElement::new(DIGEST_SIZE as u64 + 1);
         } else {
-            state[DIGEST_SIZE + 1] = BaseElement::new(value / BaseElement::MODULUS);
-            state[STATE_WIDTH - 1] = BaseElement::new(DIGEST_SIZE as u64 + 2);
+            state[INPUT2_RANGE.start + 1] = BaseElement::new(value / BaseElement::MODULUS);
+            state[CAPACITY_RANGE.start] = BaseElement::new(DIGEST_SIZE as u64 + 2);
         }
 
         // apply the Rescue permutation and return the first four elements of the state
         Self::apply_permutation(&mut state);
-        ElementDigest::new(state[..DIGEST_SIZE].try_into().unwrap())
+        ElementDigest::new(state[DIGEST_RANGE].try_into().unwrap())
     }
 }
 
@@ -198,14 +217,14 @@ impl ElementHasher for Rp64_256 {
         // is set to the number of elements to be hashed. this is done so that adding zero elements
         // at the end of the list always results in a different hash.
         let mut state = [BaseElement::ZERO; STATE_WIDTH];
-        state[STATE_WIDTH - 1] = BaseElement::new(elements.len() as u64);
+        state[CAPACITY_RANGE.start] = BaseElement::new(elements.len() as u64);
 
         // absorb elements into the state one by one until the rate portion of the state is filled
         // up; then apply the Rescue permutation and start absorbing again; repeat until all
         // elements have been absorbed
         let mut i = 0;
         for &element in elements.iter() {
-            state[i] += element;
+            state[RATE_RANGE.start + i] += element;
             i += 1;
             if i % RATE_WIDTH == 0 {
                 Self::apply_permutation(&mut state);
@@ -222,7 +241,7 @@ impl ElementHasher for Rp64_256 {
         }
 
         // return the first 4 elements of the state as hash result
-        ElementDigest::new(state[..DIGEST_SIZE].try_into().unwrap())
+        ElementDigest::new(state[DIGEST_RANGE].try_into().unwrap())
     }
 }
 
@@ -240,6 +259,27 @@ impl Rp64_256 {
     /// the remaining 4 elements are reserved for capacity.
     pub const STATE_WIDTH: usize = STATE_WIDTH;
 
+    /// The rate portion of the state is located in elements 4 through 11 (inclusive).
+    pub const RATE_RANGE: Range<usize> = RATE_RANGE;
+
+    /// The capacity portion of the state is located in elements 0, 1, 2, and 3.
+    pub const CAPACITY_RANGE: Range<usize> = CAPACITY_RANGE;
+
+    /// The output of the hash function can be read from state elements 4, 5, 6, and 7.
+    pub const DIGEST_RANGE: Range<usize> = DIGEST_RANGE;
+
+    /// MDS matrix used for computing the linear layer in a Rescue Prime round.
+    pub const MDS: [[BaseElement; STATE_WIDTH]; STATE_WIDTH] = MDS;
+
+    /// Inverse of the MDS matrix.
+    pub const INV_MDS: [[BaseElement; STATE_WIDTH]; STATE_WIDTH] = INV_MDS;
+
+    /// Round constants added to the hasher state in the first half of the Rescue Prime round.
+    pub const ARK1: [[BaseElement; STATE_WIDTH]; NUM_ROUNDS] = ARK1;
+
+    /// Round constants added to the hasher state in the second half of the Rescue Prime round.
+    pub const ARK2: [[BaseElement; STATE_WIDTH]; NUM_ROUNDS] = ARK2;
+
     // RESCUE PERMUTATION
     // --------------------------------------------------------------------------------------------
 
@@ -506,6 +546,180 @@ const MDS: [[BaseElement; STATE_WIDTH]; STATE_WIDTH] = [
     ],
 ];
 
+/// Rescue Inverse MDS matrix
+/// Computed using algorithm 4 from <https://eprint.iacr.org/2020/1143.pdf> and then
+/// inverting the resulting matrix.
+const INV_MDS: [[BaseElement; STATE_WIDTH]; STATE_WIDTH] = [
+    [
+        BaseElement::new(1025714968950054217),
+        BaseElement::new(2820417286206414279),
+        BaseElement::new(4993698564949207576),
+        BaseElement::new(12970218763715480197),
+        BaseElement::new(15096702659601816313),
+        BaseElement::new(5737881372597660297),
+        BaseElement::new(13327263231927089804),
+        BaseElement::new(4564252978131632277),
+        BaseElement::new(16119054824480892382),
+        BaseElement::new(6613927186172915989),
+        BaseElement::new(6454498710731601655),
+        BaseElement::new(2510089799608156620),
+    ],
+    [
+        BaseElement::new(14311337779007263575),
+        BaseElement::new(10306799626523962951),
+        BaseElement::new(7776331823117795156),
+        BaseElement::new(4922212921326569206),
+        BaseElement::new(8669179866856828412),
+        BaseElement::new(936244772485171410),
+        BaseElement::new(4077406078785759791),
+        BaseElement::new(2938383611938168107),
+        BaseElement::new(16650590241171797614),
+        BaseElement::new(16578411244849432284),
+        BaseElement::new(17600191004694808340),
+        BaseElement::new(5913375445729949081),
+    ],
+    [
+        BaseElement::new(13640353831792923980),
+        BaseElement::new(1583879644687006251),
+        BaseElement::new(17678309436940389401),
+        BaseElement::new(6793918274289159258),
+        BaseElement::new(3594897835134355282),
+        BaseElement::new(2158539885379341689),
+        BaseElement::new(12473871986506720374),
+        BaseElement::new(14874332242561185932),
+        BaseElement::new(16402478875851979683),
+        BaseElement::new(9893468322166516227),
+        BaseElement::new(8142413325661539529),
+        BaseElement::new(3444000755516388321),
+    ],
+    [
+        BaseElement::new(14009777257506018221),
+        BaseElement::new(18218829733847178457),
+        BaseElement::new(11151899210182873569),
+        BaseElement::new(14653120475631972171),
+        BaseElement::new(9591156713922565586),
+        BaseElement::new(16622517275046324812),
+        BaseElement::new(3958136700677573712),
+        BaseElement::new(2193274161734965529),
+        BaseElement::new(15125079516929063010),
+        BaseElement::new(3648852869044193741),
+        BaseElement::new(4405494440143722315),
+        BaseElement::new(15549070131235639125),
+    ],
+    [
+        BaseElement::new(14324333194410783741),
+        BaseElement::new(12565645879378458115),
+        BaseElement::new(4028590290335558535),
+        BaseElement::new(17936155181893467294),
+        BaseElement::new(1833939650657097992),
+        BaseElement::new(14310984655970610026),
+        BaseElement::new(4701042357351086687),
+        BaseElement::new(1226379890265418475),
+        BaseElement::new(2550212856624409740),
+        BaseElement::new(5670703442709406167),
+        BaseElement::new(3281485106506301394),
+        BaseElement::new(9804247840970323440),
+    ],
+    [
+        BaseElement::new(7778523590474814059),
+        BaseElement::new(7154630063229321501),
+        BaseElement::new(17790326505487126055),
+        BaseElement::new(3160574440608126866),
+        BaseElement::new(7292349907185131376),
+        BaseElement::new(1916491575080831825),
+        BaseElement::new(11523142515674812675),
+        BaseElement::new(2162357063341827157),
+        BaseElement::new(6650415936886875699),
+        BaseElement::new(11522955632464608509),
+        BaseElement::new(16740856792338897018),
+        BaseElement::new(16987840393715133187),
+    ],
+    [
+        BaseElement::new(14499296811525152023),
+        BaseElement::new(118549270069446537),
+        BaseElement::new(3041471724857448013),
+        BaseElement::new(3827228106225598612),
+        BaseElement::new(2081369067662751050),
+        BaseElement::new(15406142490454329462),
+        BaseElement::new(8943531526276617760),
+        BaseElement::new(3545513411057560337),
+        BaseElement::new(11433277564645295966),
+        BaseElement::new(9558995950666358829),
+        BaseElement::new(7443251815414752292),
+        BaseElement::new(12335092608217610725),
+    ],
+    [
+        BaseElement::new(184304165023253232),
+        BaseElement::new(11596940249585433199),
+        BaseElement::new(18170668175083122019),
+        BaseElement::new(8318891703682569182),
+        BaseElement::new(4387895409295967519),
+        BaseElement::new(14599228871586336059),
+        BaseElement::new(2861651216488619239),
+        BaseElement::new(567601091253927304),
+        BaseElement::new(10135289435539766316),
+        BaseElement::new(14905738261734377063),
+        BaseElement::new(3345637344934149303),
+        BaseElement::new(3159874422865401171),
+    ],
+    [
+        BaseElement::new(1134458872778032479),
+        BaseElement::new(4102035717681749376),
+        BaseElement::new(14030271225872148070),
+        BaseElement::new(10312336662487337312),
+        BaseElement::new(12938229830489392977),
+        BaseElement::new(17758804398255988457),
+        BaseElement::new(15482323580054918356),
+        BaseElement::new(1010277923244261213),
+        BaseElement::new(12904552397519353856),
+        BaseElement::new(5073478003078459047),
+        BaseElement::new(11514678194579805863),
+        BaseElement::new(4419017610446058921),
+    ],
+    [
+        BaseElement::new(2916054498252226520),
+        BaseElement::new(9880379926449218161),
+        BaseElement::new(15314650755395914465),
+        BaseElement::new(8335514387550394159),
+        BaseElement::new(8955267746483690029),
+        BaseElement::new(16353914237438359160),
+        BaseElement::new(4173425891602463552),
+        BaseElement::new(14892581052359168234),
+        BaseElement::new(17561678290843148035),
+        BaseElement::new(7292975356887551984),
+        BaseElement::new(18039512759118984712),
+        BaseElement::new(5411253583520971237),
+    ],
+    [
+        BaseElement::new(9848042270158364544),
+        BaseElement::new(809689769037458603),
+        BaseElement::new(5884047526712050760),
+        BaseElement::new(12956871945669043745),
+        BaseElement::new(14265127496637532237),
+        BaseElement::new(6211568220597222123),
+        BaseElement::new(678544061771515015),
+        BaseElement::new(16295989318674734123),
+        BaseElement::new(11782767968925152203),
+        BaseElement::new(1359397660819991739),
+        BaseElement::new(16148400912425385689),
+        BaseElement::new(14440017265059055146),
+    ],
+    [
+        BaseElement::new(1634272668217219807),
+        BaseElement::new(16290589064070324125),
+        BaseElement::new(5311838222680798126),
+        BaseElement::new(15044064140936894715),
+        BaseElement::new(15775025788428030421),
+        BaseElement::new(12586374713559327349),
+        BaseElement::new(8118943473454062014),
+        BaseElement::new(13223746794660766349),
+        BaseElement::new(13059674280609257192),
+        BaseElement::new(16605443174349648289),
+        BaseElement::new(13586971219878687822),
+        BaseElement::new(16337009014471658360),
+    ],
+];
+
 // ROUND CONSTANTS
 // ================================================================================================
 
